#import "AVAudio_Tap.h"

@implementation AVAudio_Tap

-(id)       init {
    
    audioProcessor = [[AudioProcessor alloc] init];

    bpmUpdateTimer = dispatch_source_create (DISPATCH_SOURCE_TYPE_TIMER, 0, 0, dispatchQueue);
    
    dispatch_source_set_timer (bpmUpdateTimer, DISPATCH_TIME_NOW, 1.0 * NSEC_PER_SEC, 0.0);
    
    dispatch_source_set_event_handler (bpmUpdateTimer, ^{
        
        int bpm = [self->audioProcessor getBPM];                        // Simplify tempo to an integer value
        
        if (bpm >= MIN_BPM && bpm <= MAX_BPM) {                         // Only remember range-limited tempos
            
            if (bpm != self->currentBPM) {
                
                NSLog (@"Tempo changed (from %d to %d BPM)", self->currentBPM, bpm);

                self->currentBPM = bpm;
            }
            
        } else {
            
            NSLog (@"Tempo ignored (Out of Range: %d BPM)", bpm);
            
            self->currentBPM = 0;
        }
    });

    return self;
}

-(void)     startTap {
    
    [audioProcessor start];
    
    dispatch_activate (bpmUpdateTimer);
}

-(void)     endTap {
    
    if (bpmUpdateTimer) {
        
        dispatch_source_cancel (bpmUpdateTimer);
        
        [audioProcessor stop];
        
        bpmUpdateTimer = nil;
        
        audioProcessor = nil;
    }
}

-(float)    getCurrentBPM { return currentBPM; }

-(void)     dealloc {
    
    [self endTap];
}

@end
