#ifndef _BVH_
#define _BVH_

#include <vector>
#include "bvh_math.h"

class BVH {
    
public:
    
    enum Channel { Xpos=1, Ypos, Zpos, Xrot, Yrot, Zrot };
    
    struct Part {
        
        int                  index;
        int                  parent;
        int                  channels;
        int                  childCount;
        std::vector<int>     childIndices;
        char*                name;
        BVH_Math::vec3       offset;
        BVH_Math::vec3       end;
        BVH_Math::Transform* motion;
    };
    
public:
    
    BVH();
    ~BVH();
    
    bool        load (const char *data);
    
    const Part* getPart (int index) const   { return m_parts[index]; }
    int         getPartCount() const        { return m_partCount; }
    int         getFrameCount() const       { return m_frameCount; }
    float       getFrameTime() const        { return m_frameTime; }

private:
    
    Part*       readHeirachy (const char*& data);
    
protected:
    
    Part*   m_root;
    Part**  m_parts;
    int     m_partCount;
    int     m_frameCount;
    float   m_frameTime;
};

#endif
