#import "Controller.h"

@implementation Controller

-(void) awakeFromNib {
    
    [NSApp setDelegate  :(id <NSApplicationDelegate>) self];
    
    [glWindow setCollectionBehavior :(NSWindowCollectionBehaviorFullScreenPrimary)];
    
    [glWindow makeFirstResponder  :self];
    
    P2_Logging_Enabled = false;
    
    glView = [[View alloc] initWithFrame  :[glWindow frame]  :P2_Logging_Enabled];

    if	(glView != nil) {
        
        [glWindow setContentView  :glView];
        
        [glWindow makeKeyAndOrderFront  :self];
        
        [glView setMesh  :[glView loadP2MeshFromBundle  :'6']];
        
        animations[typeArmature] = animations[typeAction] = animations[typeBVH]  = nil;
        
        [self meshReadout  :false  :false];
        
        [self setupRenderTimer];
        
        oldTempo = 0.0;
        
    } else {
        
        NSAlert *alert = [NSAlert new];
        
        [alert setAlertStyle       :NSAlertStyleWarning];
        [alert setMessageText      :@"Failed to initialize OpenGL"];
        [alert addButtonWithTitle  :@"OK"];
        
        [alert runModal];
        
        [MeshP2 P2Log  :P2_Logging_Enabled  :@"DisplayP2->awakeFromNib: Failed to initialize OpenGL"];
        
        [NSApp terminate  :self];
    }
}

-(void) setupRenderTimer {
    
    NSTimeInterval timeInterval = 1/60;
    
    if (renderTimer != nil && [renderTimer isValid]) [renderTimer invalidate];
    
    renderTimer = [NSTimer scheduledTimerWithTimeInterval  :timeInterval target:self selector:@selector(updateGLView:) userInfo:nil repeats:YES];
    
    [[NSRunLoop currentRunLoop] addTimer  :renderTimer forMode:NSDefaultRunLoopMode];
}

-(void) setupTempoTimer {
    
    NSTimeInterval timeInterval = 1.0;
    
    if (tempoTimer != nil && [tempoTimer isValid]) [tempoTimer invalidate];
    
    tempoTimer = [NSTimer scheduledTimerWithTimeInterval  :timeInterval target:self selector:@selector(updateTempo:) userInfo:nil repeats:YES];
    
    [[NSRunLoop currentRunLoop] addTimer  :tempoTimer forMode:NSDefaultRunLoopMode];
}

-(void) updateGLView  :(NSTimer*)timer { if (glView != nil) [glView drawRect  :[glView frame]]; }

-(void) updateTempo  :(NSTimer*)timer {
    
    if (audioTap) {
        
        newTempo = [audioTap getCurrentBPM];
        
        if ((newTempo != oldTempo) && (newTempo != 0)) {
            
            animations[typeArmature] = (animations[typeArmature]) ? [self loadAnimationByType  :typeArmature  :[glView getMesh]]  :  nil;
            
            [[glView getMesh] UploadAnimation  :animations[typeArmature]  :typeArmature];  // Re-upload tempo sensitive Armature animation, if any
            
            oldTempo = newTempo;
        }
    }
}

-(void) keyDown  :(NSEvent*)theEvent {
    
    unichar unicodeKey = [[theEvent characters] characterAtIndex:0];
    
    bool numericPadKey = ([theEvent modifierFlags] & NSNumericPadKeyMask) != 0;     // Numeric keypad pressed
    bool commandKey    = ([theEvent modifierFlags] & NSCommandKeyMask) != 0;        // Command key (i.e. '⌘')
    
    switch (unicodeKey) {
            
        case '+':                                                           // '+' key (i.e. shifted or not)
        case '=':
            
            [[glView getOkudagram]  SetHyperspaceLocked  :true];            // Lock Hyperspace effect on
            break;
            
        case '-':
            
            [[glView getOkudagram]  SetHyperspaceLocked  :false];           // Unlock Hyperspace effect
            break;
            
        case '0':                                                           // Handle number keys and/or keypad keys
        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':
        case '9':
            
            if (numericPadKey) {
                
                [[glView getOkudagram] SetWarpSpeed  :(GLfloat)(unicodeKey - '0')]; // Convert unicodeKey to binary

                [glView initCameraLightingAndMatrices  :true];                      // Warp speeds from 0 to 9
                
            } else {
                
                animations[typeArmature] = animations[typeAction] = animations[typeBVH]  = nil;
                
                [glView setMesh  :[glView loadP2MeshFromBundle  :unicodeKey]];      // Select mesh by bundle ID
            }

            break;
            
        case 'a':                                                           // Toggles Armature or Action animations on/off
        case 'A': {
            
            uint boneCount   = (uint)[[[glView getMesh] getArmatureList] count];
            uint actionCount = (uint)[[[glView getMesh] getActionList] count];
            
            [MeshP2 P2Log  :P2_Logging_Enabled  :@"DisplayP2->keyDown: Animation requested for %d Armature bones and/or %d Actions", boneCount, actionCount];
            
            if (boneCount || actionCount) {
                
                if (glView != nil) {

                    if ((boneCount) && !(commandKey)) {                     // Default simply toggles any available Armature animation
                        
                        animations[typeAction]   = nil;                     // Turn off Action animation handling as it's mutually exclusive with Armature
                        
                        animations[typeArmature] = (animations[typeArmature]) ? nil
                                                                              : [self loadAnimationByType  :typeArmature  :[glView getMesh]];
                        
                        [[glView getMesh] UploadAnimation  :animations[typeArmature]  :typeArmature];
                    }
                    
                    if ((actionCount) && (commandKey)) {                    // ⌘-A key toggles Action animation (e.g. "Loading..." Splash screen)

                        animations[typeArmature] = nil;                     // Turn off Armature animation handling as it's mutually exclusive with Action

                        animations[typeAction] = (animations[typeAction])     ? nil
                                                                              : [self loadAnimationByType  :typeAction  :[glView getMesh]];
                        
                        [[glView getMesh] UploadAnimation  :animations[typeAction]  :typeAction];
                    }
                }
                
            } else {
                
                NSAlert *alert = [NSAlert new];
                
                [alert setAlertStyle       :NSAlertStyleWarning];
                [alert setInformativeText  :@"This mesh has neither an Action nor Armature Bones loaded."];
                [alert setMessageText      :@"Animation Failure"];
                [alert addButtonWithTitle  :@"OK"];
                [alert runModal];
            }
            
            break;
        }
            
        case 'b':                                                           // Toggle BVH Animation on/off
        case 'B': {

            if (glView != nil) {
                
                if (commandKey) {                                           // Find a BVH file matching the mesh file name
                    
                    if ((animations[typeBVH] = [self loadBVHFileMatchingMeshFile  :[[glView getMesh] getMeshFilename]])) {
                        
                        [[glView getMesh] UploadAnimation  :animations[typeBVH]  :typeBVH];
                    }

                } else {                                                    // Default toggles or prompts a BVH animation
                    
                    if (animations[typeBVH]) {                              // If there is a BVH animation already running
                        
                        animations[typeBVH] = nil;                          // Invalidate the current BVH animation status

                        [[glView getMesh] UploadAnimation  :nil  :typeBVH]; // Notify MeshP2 the BVH animation is invalid
                        
                    } else {
                        
                        if ((animations[typeBVH] = [self openBVHFileCallback])) {                   // Select, load .BVH file

                            uint jointCount = (uint)[[[glView getMesh] getBVHList] count];          // Fetch mesh joint count
                            
                            if (!jointCount) break;                                                 // Bail if Jointless mesh

                            if (jointCount == [[animations[typeBVH] getBoneAnimations] count]) {    // An extra safety check
                                
                                [[glView getMesh] UploadAnimation  :animations[typeBVH]  :typeBVH]; // Upload BVH animations
                                
                                [MeshP2 P2Log  :P2_Logging_Enabled  :@"DisplayP2->keyDown: BVH Animation requested for %d joints", jointCount];

                            } else {
                                
                                NSAlert *alert = [NSAlert new];
                                
                                [alert setAlertStyle       :NSAlertStyleWarning];
                                [alert setInformativeText  :@"The selected BVH file's Joint count doesn't match the loaded mesh's Bone count."];
                                [alert setMessageText      :@"Please select a compatible BVH file"];
                                [alert addButtonWithTitle  :@"OK"];
                                [alert runModal];
                                
                                animations[typeBVH] = nil;                  // Forget this borked BVH animation altogether

                                break;
                            }
                        }
                    }
                }
            }
            
            break;
        }
            
        case 'c':                                                           // Apply a Blender->OpenGL axis 'Correction':
        case 'C': {                                                         // Viewed from origin, rotate -90° on X-axis,
                                                                            // to display/animate the mesh, Blender style
            if (glView != nil) {
                
                MeshP2   *mesh = [glView getMesh];                          // Save a little object de-referencing overhead

                glm::mat4 glmCorrectionMatrix = glm::rotate (glm::mat4 (1.0), glm::radians (-90.0f), glm::vec3 (1.0, 0.0, 0.0));
                
                Matrix4f *correctionMatrix    = [[Matrix4f new] cmMatrixFromBytes  :glm::value_ptr (glmCorrectionMatrix)];
                
                          correctionMatrix    = [Matrix4f multiply  :[mesh getModelViewMatrix]  :correctionMatrix];
                
                [mesh SetModelViewMatrix  :correctionMatrix];               // Use 'r' key to Reset excess Corrections
            }
            
            break;
        }
            
        case 'd':                                                           // Double mesh size (halve model distance)
        case 'D': {
            
            glm::vec3 ncp = [glView getCameraPosition] * glm::vec3 (0.5);   // Calculate potential New Camera Position
            
            float limit = std::max ({abs(ncp.x), abs(ncp.y), abs(ncp.z)});  // Camera's highest value seen on any axis
            
            if (limit < FAR_CLIPPING_PLANE) {                               // Ensure it is within the viewing frustum
                
                [glView setCameraPosition  :ncp];                           // If so, use it. Otherwise, just ignore it
                
                [glView initCameraLightingAndMatrices  :false];             // Reset lighting & matrices with new camera
            }
            
            break;
        }
            
        case 'f':                                                           // Fly-by animation
        case 'F':
            
            [glView setFlyBy_Enabled  :![glView getFlyBy_Enabled]];
            
            break;
            
        case 'h':                                                           // Halve mesh size (double camera->mesh distance)
        case 'H': {
            
            glm::vec3 ncp = [glView getCameraPosition] * glm::vec3 (2.0);   // Calculate potential New Camera Position
            
            float limit = std::max ({abs(ncp.x), abs(ncp.y), abs(ncp.z)});  // Camera's highest value seen on any axis
            
            if (limit < FAR_CLIPPING_PLANE) {                               // Ensure it is within the viewing frustum
                
                [glView setCameraPosition  :ncp];                           // If so, use it. Otherwise, just ignore it
                
                [glView initCameraLightingAndMatrices  :false];             // Reset lighting & matrices with new camera
            }
            
            break;
        }
            
        case 'l':                                                           // Start listening to microphone, process tempo
        case 'L': {
            
            uint boneCount   = (uint)[[[glView getMesh] getArmatureList] count];
            uint actionCount = (uint)[[[glView getMesh] getActionList] count];
            uint bvhCount    = (uint)[[[glView getMesh] getBVHList] count];

            if (boneCount || actionCount || bvhCount) {
                
                if (glView != nil) {
                    
                    [self enableAudioTap];
                    
                    [self setupTempoTimer];
                }
                
            } else {
                
                NSAlert *alert = [NSAlert new];
                
                [alert setAlertStyle       :NSAlertStyleWarning];
                [alert setInformativeText  :@"This mesh lacks skeletal bones (neither Action, Armature nor BVH)"];
                [alert setMessageText      :@"Animation Failure"];
                [alert addButtonWithTitle  :@"OK"];
                [alert runModal];
            }
            
            break;
        }
            
        case 'm':                                                           // Select 3D deep field skymap
        case 'M':
            
            if (!commandKey) [[glView getOkudagram] SetSkymap  :0];          // m-key selects primary skymap
            else             [[glView getOkudagram] SetSkymap  :1];          // ⌘-m selects secondary skymap
            break;
            
        case 'o':                                                           // Toggle ocean wave effect on/off
        case 'O':
            
            [[glView getOkudagram] ToggleOcean];
            break;
            
        case 'r':
        case 'R': {
            
            [glView initCameraLightingAndMatrices  :true];                  // Init the View's camera, lighting, MVPs & view frustum

            [glView setFlyBy_Enabled  :false];
            
            break;
        }
            
        case 's':                                                           // Toggle 3D deep field skymap on/off
        case 'S':
            
            [[glView getOkudagram] ToggleSkymap];
            break;
            
        case 'w':
        case 'W':

            if (!commandKey) [[glView getOkudagram] ToggleWireframeMode];   // w-key toggles wireframe mode for all polygons
            else             [[glView getOkudagram] ToggleWarpfield];       // ⌘-w toggles animated warpfield effect on/off
            break;
            
        case 27:                                                            // ESC/Q/X key exits at top of event loop
        case 'q':
        case 'Q':
        case 'x':
        case 'X':
            
            [NSApp performSelector  :@selector(terminate:)  withObject:nil  afterDelay:0.0];
            
            break;
    }
}

-(void) enableAudioTap {
    
    if (@available(macOS 10.14, *)) {
        
        switch ([AVCaptureDevice authorizationStatusForMediaType  :AVMediaTypeAudio]) {
                
            case AVAuthorizationStatusAuthorized: {
                
                if (audioTap) [audioTap endTap];                            // Clear any previously created audioTaps
                
                audioTap = [AVAudio_Tap new];
                
                [audioTap startTap];
                
                break;
            }
                
            case AVAuthorizationStatusNotDetermined: {
                
                [AVCaptureDevice requestAccessForMediaType  :AVMediaTypeAudio completionHandler:^(BOOL granted) {
                    
                    if (granted) {
                        
                        dispatch_async (dispatch_get_main_queue(), ^{
                            
                            [self enableAudioTap];
                        });
                    }
                }];
                
                break;
            }
                
            case AVAuthorizationStatusDenied: {
                
                NSString *informativeText= [NSString stringWithFormat  :@"Re-enable in System Preferences:\n\t->Security & Privacy\n\t->Privacy tab\n\t->Microphone\n\t->DisplayP2"];
                
                NSAlert *alert = [NSAlert new];
                
                [alert setAlertStyle       :NSAlertStyleWarning];
                [alert setInformativeText  :informativeText];
                [alert setMessageText      :@"Microphone Access is currently Disabled"];
                [alert addButtonWithTitle  :@"OK"];
                
                [alert runModal];
                
                [MeshP2 P2Log  :P2_Logging_Enabled  :@"Microphone access has been previously denied"];
                
                break;
            }
                
            case AVAuthorizationStatusRestricted: {
                
                NSString *informativeText= [NSString stringWithFormat  :@"Your UserID is restricted from granting Microphone access."];
                
                NSAlert *alert = [NSAlert new];
                
                [alert setAlertStyle       :NSAlertStyleWarning];
                [alert setInformativeText  :informativeText];
                [alert setMessageText      :@"Microphone Access Restricted"];
                [alert addButtonWithTitle  :@"OK"];
                [alert runModal];
                
                [MeshP2 P2Log  :P2_Logging_Enabled  :@"UserID is restricted from granting Microphone access."];
                
                break;
            }
        }
    }
}

-(IBAction) openP2MeshFile  :(id)sender {
    
    NSArray *fileTypes = [NSArray arrayWithObjects  :@"p2", nil];
    
    NSOpenPanel *oPanel = [NSOpenPanel openPanel];                          // Create the File Open Panel class
    
    [oPanel setAlphaValue               :0.95];                             // Specify panel's Alpha value (i.e. transparency)
    [oPanel setCanChooseFiles           :YES];                              // Enable ability to select a file from the panel
    [oPanel setCanChooseDirectories     :NO];                               // Disable selection of directories in the panel
    [oPanel setCanCreateDirectories     :NO];                               // Disable creation of directories in the panel
    [oPanel setAllowsMultipleSelection  :NO];                               // Disable the ability to select multiple files
    [oPanel setAllowedFileTypes         :fileTypes];                        // Specify pre-selection of .P2 format meshes
    [oPanel setTitle                    :@"Select a .P2 mesh file"];
    
    if ([oPanel runModal] == NSModalResponseOK) {
        
        NSArray *urls = [oPanel URLs];                                      // With setAllowsMultipleSelection:NO there'll only be one URL
        
        NSURL *url = [urls objectAtIndexedSubscript  :0];
        
        NSString *meshFileName = [[url resourceSpecifier] stringByRemovingPercentEncoding];
        
        NSString *errorMessageText = [NSString stringWithFormat  :@"Error opening .p2 mesh file: '%@'\n\nTry another file?", meshFileName];
        
        @try {
            
            animations[typeArmature] = animations[typeAction] = nil;        // Reset primary (i.e. non-BVH) animations
            
            [glView loadP2MeshByFileName  :meshFileName];                   // Parse and load the requested mesh file
        }
        
        @catch (NSException *e) {
            
            NSAlert *alert = [NSAlert new];
            
            [alert setAlertStyle       :NSAlertStyleWarning];
            [alert setInformativeText  :errorMessageText];
            [alert setMessageText      :@".p2 Mesh File Error!\nCheck Log for openFile backtrace."];
            [alert addButtonWithTitle  :@"OK"];
            
            [alert runModal];
            
            NSArray *backtrace = [e callStackSymbols];
            
            [MeshP2 P2Log  :P2_Logging_Enabled  :@"Exception raised in openFile e = %@.\n\n Backtrace:\n%@", e, backtrace];
            
            return;                                                         // No point continuing
        }
        
        [self meshReadout  :false  :false];                                 // Read out some stats
    }
}

-(IBAction) openBVHFile  :(id)sender {
    
    uint boneCount = (uint)[[[glView getMesh] getArmatureList] count];

    if ((animations[typeBVH] = [self openBVHFileCallback])) {
        
        if (boneCount == [[animations[typeBVH] getBoneAnimations] count]) {         // An extra safety check
            
            [[glView getMesh] UploadAnimation  :animations[typeBVH]  :typeBVH];     // Upload BVH animation
            
        } else {
            
            NSAlert *alert = [NSAlert new];
            
            [alert setAlertStyle       :NSAlertStyleWarning];
            [alert setInformativeText  :@"The selected BVH file's bone count doesn't match the loaded mesh's bone count. Select a compatible BVH file"];
            [alert setMessageText      :@"BVH Error"];
            [alert addButtonWithTitle  :@"OK"];
            [alert runModal];
            
            animations[typeBVH] = nil;                                              // Forget this borked BVH animation altogether
        }
    }
}

-(Animation*) openBVHFileCallback {                                                 // Factored out this method to be callable from keyDown & openBVHFile

    NSArray *fileTypes = [NSArray arrayWithObjects  :@"bvh", nil];
    
    NSOpenPanel *oPanel = [NSOpenPanel openPanel];
    
    [oPanel setAlphaValue               :0.95];                                     // Specify panel's Alpha value (i.e. transparency)
    [oPanel setCanChooseFiles           :YES];                                      // Enable ability to select a file from the panel
    [oPanel setCanChooseDirectories     :NO];                                       // Disable selection of directories in the panel
    [oPanel setCanCreateDirectories     :NO];                                       // Disable creation of directories in the panel
    [oPanel setAllowsMultipleSelection  :NO];                                       // Disable the ability to select multiple files
    [oPanel setAllowedFileTypes         :fileTypes];                                // Specify pre-selection of .BVH format files
    [oPanel setTitle                    :@"Select a .BVH file"];
    
    if ([oPanel runModal] == NSModalResponseOK) {
        
        NSArray *urls = [oPanel URLs];                                              // With setAllowsMultipleSelection:NO there'll only be one URL
        
        NSURL *url = [urls objectAtIndexedSubscript  :0];
        
        NSString *bvhFilename = [[url resourceSpecifier] stringByRemovingPercentEncoding];

        @try {

            return [glView loadBVHanimation  :bvhFilename  :[audioTap getCurrentBPM]];;
        }
        
        @catch (NSException *e) {
            
            NSAlert *alert = [NSAlert new];
            
            [alert setAlertStyle       :NSAlertStyleWarning];
            [alert setInformativeText  :@"Check Log for BVH Parser failure"];
            [alert setMessageText      :[e reason]];
            [alert addButtonWithTitle  :@"OK"];
            
            [alert runModal];
        }
    }

    return nil;
}

-(Animation*) loadBVHFileMatchingMeshFile  :(NSString*)in_MeshFilename {
    
    NSString *bvhFilename = [[in_MeshFilename stringByDeletingPathExtension] stringByAppendingString  :@".bvh"];    // Replace file extension with .bvh
    
    NSString *errorMessageText = [NSString stringWithFormat  :@"Try 'File->Apply .BVH Animation...' manually?"];
    
    @try {
        
        return [glView loadBVHanimation  :bvhFilename  :[audioTap getCurrentBPM]];                      // Parse and load the matching BVH animation;
    }
    
    @catch (NSException *exception) {
        
        animations[typeBVH] = nil;                                                                      // Stop any residual BVH animations

        NSAlert *alert = [NSAlert new];
        
        [alert setAlertStyle       :NSAlertStyleWarning];
        [alert setInformativeText  :errorMessageText];
        [alert setMessageText      :exception.reason];
        [alert addButtonWithTitle  :@"OK"];
        
        [alert runModal];
        
        return nil;
    }
}

-(Animation*) loadAnimationByType  :(uint)in_AnimType  :(MeshP2*)in_Mesh {                              // Returns a beat-driven Animation object
    
    NSMutableArray *newAnimationKeyframes = [NSMutableArray new];                                       // Resultant Animation keyframes array
    NSMutableArray *newAnimationDurations = [NSMutableArray new];                                       //     "         "     durations   "
    
    float   tempo, frameTime, lastFrameTime;
    
    if (in_AnimType == typeArmature) {                                                                  // If this is an Armature-based anim
        
        for (id bone in [in_Mesh getArmatureList]) {                                                    // For each BONE object in ArmatureList

            NSArray        *poses         = [bone getAnimations];                                       // Get imported (Local-space) animations
            NSUInteger      nPoses        = [poses count];                                              // Number of poses imported for this bone
            NSMutableArray *boneKeyFrames = [NSMutableArray arrayWithCapacity  :nPoses];                // Holds array of keyframes for this bone

            lastFrameTime   = 0.0;                                                                      // Computed timestamp for pose iterations

            for (int poseIndex = 0; poseIndex < nPoses; poseIndex++) {                                  // For each of this bone's available poses:

                tempo         = [audioTap getCurrentBPM];                                               // Get currently detected tempo, if any
                
                frameTime     = (tempo > 0.0) ? tempo : 120.0;                                          // If none detected, default to 120 BPM
                
                frameTime     = ((frameTime / 60.0) * (1.0 / 24.0) / 2.0);                              // Match frame time to detected (half beat) tempo

                lastFrameTime = poseIndex * frameTime;                                                  // Timestamp is multiple of frameTime or 24FPS slice
                
                Matrix4f *poseMatrix = poses[poseIndex];                                                // Get imported animated pose transform at this pose
                
                DualQuaternion *poseDQ = [[DualQuaternion alloc] initFromMatrix  :poseMatrix];          // Convert pose from Matrix4f to DualQuaternion format
                
                BoneKeyFrame *keyframe = [[BoneKeyFrame alloc] init  :lastFrameTime  :poseDQ];          // Create timestamped BoneKeyFrame from the current pose
                
                [boneKeyFrames addObject  :keyframe];                                                   // Add timestamped BoneKeyFrame to array of bone's keyframes
            }
            
            [newAnimationKeyframes addObject  :boneKeyFrames];                                          // Add array of bone's keyframes to array of bone's animations
            
            [newAnimationDurations addObject  :[NSNumber numberWithFloat:lastFrameTime]];               // Add duration of bone's keyframes to array of bone's durations
        }
        
    } else if (in_AnimType == typeAction) {                                                             // Else, if an Action-based animation
        
        for (id action in [in_Mesh getActionList]) {                                                    // For each ACTION object in ActionList:
            
            NSArray        *poses         = [action getAnimations];                                     // Get imported (Model-space) animations
            NSUInteger      nPoses        = [poses count];                                              // Number of poses imported for this bone
            NSMutableArray *boneKeyFrames = [NSMutableArray arrayWithCapacity  :nPoses];                // Holds array of keyframes for this bone

            lastFrameTime   = 0.0;                                                                      // Computed timestamp for pose iterations

            for (int poseIndex = 0; poseIndex < nPoses; poseIndex++) {                                  // For each of this bone's available poses:
                
                tempo         = [audioTap getCurrentBPM];                                               // Get currently detected tempo, if any
                
                frameTime     = (tempo > 0.0) ? tempo : 120.0;                                          // If none detected, default to 120 BPM
                
                frameTime     = ((frameTime / 60.0) * (1.0 / 24.0) / 2.0);                              // Match frame time to detected (half beat) tempo

                lastFrameTime = poseIndex * frameTime;                                                  // Timestamp is multiple of frameTime or 24FPS slice
                
                Matrix4f *poseMatrix = poses[poseIndex];                                                // Get imported animated pose transform at this pose
                
                DualQuaternion *dqPoseTransform = [[DualQuaternion alloc] initFromMatrix  :poseMatrix]; // Convert pose from matrix to DualQuaternion format
                
                BoneKeyFrame *currentKeyFrame = [[BoneKeyFrame alloc] init  :lastFrameTime              // Create timestamped BoneKeyFrame from the current pose
                                                                            :dqPoseTransform];
                [boneKeyFrames addObject  :currentKeyFrame];                                            // Add timestamped BoneKeyFrame to array of bone's keyframes
            }
            
            [newAnimationKeyframes addObject  :boneKeyFrames];                                          // Add array of bone's keyframes to array of bone's animations
            
            [newAnimationDurations addObject  :[NSNumber numberWithFloat  :lastFrameTime]];             // Add duration of bone's keyframes to array of bone's durations
        }
    }
    
    return [[Animation alloc] init  :newAnimationKeyframes  :newAnimationDurations];                    // Otherwise, return a new (and possibly empty) Animation object
}

-(void) meshReadout  :(bool)readout_Vertices  :(bool)readout_Bones {
    
    if (readout_Vertices) {
        
        [[[glView getMesh] getVertexList] enumerateObjectsUsingBlock  :^(VERTEX *vertex, NSUInteger index, BOOL *stop) {
            
            NSMutableArray *flags = [vertex getFlags];                      // For this vertex, get any flags
            
            for (int f=0; f < [flags count]; f++) {                         // For each available vertex flag:
                
                uint16 flag = [flags[f] unsignedShortValue];                // Dereference the vertex flag element
                
                if (flag != 0xFFFF) {                                       // Ignore initialized placeholder flags
                    
                    uint16 BoneIndex = ((flag & 0xFF80) >> 7);              // Isolate Bone Index (9 bits from 7-15)
                    
                    uint8  VertexWeight  =  (flag & 0x007F);                // Isolate Vertex Weight (7 bits from 0-6)
                    
                    [MeshP2 P2Log  :P2_Logging_Enabled  :@"DisplayP2->openFile: VertexIndex:%lu xyz:(%.3f, %.3f, %.3f), BoneIndex:%u, VertexWeight:%3u%%",
                           (unsigned long)index, vertex.getX, vertex.getY, vertex.getZ, BoneIndex, VertexWeight];
                }
            }
        }];
    }
    
    if (readout_Bones) {
        
        [[[glView getMesh] getArmatureList] enumerateObjectsUsingBlock  :^(BONE *bone, NSUInteger index, BOOL *stop) {
            
            assert (index == [bone getIndex]);                              // Verify Blender bone index matches
            
            Vector3f *location = [bone getLocation];
            Vector4f *rotation = [bone getRotation];
            Vector3f *scale    = [bone getScale];
            
            [MeshP2 P2Log  :P2_Logging_Enabled  :@"DisplayP2->openFile: Bone:%u, Parent:%u '%@'\t"
                   @"Children:%u  "
                   @"Location:(%1.3f, %1.3f, %1.3f)  "
                   @"Rotation:(%1.3f, %1.3f, %1.3f, w:%1.3f)  "
                   @"Scale:(%1.3f, %1.3f, %1.3f)",
                   [bone getIndex], [bone getParentIndex], [bone getBoneName],
                   (unsigned short)[[bone getChildIndices] count],
                   [location getX], [location getY], [location getZ],
                   [rotation getX], [rotation getY], [rotation getZ], [rotation getW],
                   [scale getX], [scale getY], [scale getZ]];
        }];
    }
}

-(void) application  :(NSApplication*)sender openFile  :(NSString*)in_fileName {
    
// See: https://stackoverflow.com/questions/38272650/cocoa-app-open-file-via-double-click-or-right-click
    
    if  (glView) {
        
        [glView loadP2MeshByFileName  :in_fileName];
        
        animations[typeArmature] = animations[typeAction] = animations[typeBVH] = nil;  // Stop any animations
    }
}

-(BOOL) applicationSupportsSecureRestorableState  :(NSApplication*)app { return YES; }  // Silence warning

-(void) dealloc {
    
    if (renderTimer != nil && [renderTimer isValid]) [renderTimer invalidate];
    
    if (tempoTimer != nil && [tempoTimer isValid]) [tempoTimer invalidate];
}

@end
