import os
import bvhio

source      = "/Users/pinter/Desktop/Bear/Bear_13_18.bvh"
destination = "/Users/pinter/Desktop/Bear/Bear_13_18_Corrected.bvh"

print (f"Loading {os.path.basename (source)}")

bvhRoot         = bvhio.readAsBvh (source).Root

hierarchyRoot   = bvhio.convertBvhToHierarchy (bvhRoot)

hierarchyRoot.loadRestPose (recursive=True)

for joint in hierarchyRoot.filter ('Neck1'):

    joint.Rotation *= bvhio.Euler.toQuatFrom ((-0.785398163, 0.0, 0.0), 'ZYX')     # Rotate 1st neck bone on X-axis -45° (radians)

for joint in hierarchyRoot.filter ('Neck2'):

    joint.Rotation *= bvhio.Euler.toQuatFrom ((-0.261799388, 0.0, 0.0), 'ZYX')     # Rotate 2nd neck bone -15°

for joint in hierarchyRoot.filter ('RBLegAnkle'):

    joint.Rotation *= bvhio.Euler.toQuatFrom ((-0.785398163, 0.0, 0.0), 'ZYX')     # Rotate 1st ankle bone -45°

for joint in hierarchyRoot.filter ('LBLegAnkle'):

    joint.Rotation *= bvhio.Euler.toQuatFrom ((-0.785398163, 0.0, 0.0), 'ZYX')     # Rotate 2nd ankle bone -45°

hierarchyRoot.writeRestPose (recursive=True, keep=['position', 'rotation', 'scale'])

bvhRoot = bvhio.convertHierarchyToBvh (hierarchyRoot, hierarchyRoot.getKeyframeRange()[1] + 1)

for joint, index, depth in bvhRoot.layout():

    joint.Channels = ['Xposition', 'Yposition', 'Zposition']

    joint.Channels.extend (['Zrotation', 'Yrotation', 'Xrotation'])

print (f"Writing {os.path.basename (destination)}")

bvhio.writeBvh (destination, bvhio.BvhContainer (bvhRoot, len(bvhRoot.Keyframes), 1/24))
