#version 150 core                           // Stipulate OpenGL 3.2 Core

uniform mat4    ModelViewMatrix;            // Transforms vertex to Eye-space
uniform mat4    ProjectionMatrix;           // Transforms vertex to Clip-space
uniform mat4    TransformMatrix;            // Translates vertex onto ellipsoid
uniform vec4    AttenuatedColor;            // Distance-attenuated vertex colour

in      vec2    Position;                   // Input from VBO: Vertex's 2D position
out     vec4    Alpha;                      // Output to next stage: Vertex's alpha

void main (void) {

    Alpha = (gl_VertexID != 0) ? AttenuatedColor : vec4 (1.0, 1.0, 1.0, AttenuatedColor.a);   // Set star's origin vertex (v0) to white
    
    vec4 newPosition = TransformMatrix * vec4 (Position, 0.0, 1.0); // Translate vertex's 2D VBO position onto the 3D ellipsoid surface

    gl_Position = ProjectionMatrix * ModelViewMatrix * newPosition; // Transform position on ellipsoid to Eye-space, then to Clip-space
}
