#define GL_DO_NOT_WARN_IF_MULTI_GL_VERSION_HEADERS_INCLUDED

#import <ScreenSaver/ScreenSaver.h>
#import <OpenGL/gl.h>
#import "Okudagram.h"

#define DEF_STARS        750                            // Default number of stars in 3D starfield
#define DEF_COLOR        (glm::vec3 (0.3, 0.3, 1.0))    // Default RGB colour of stars in starfield
#define DEF_SPEED        5.0                            // Default warp speed for animated 3D stars
#define MIN_SPEED        0.000001                       // Minimum speed used to limit side effects
#define MAX_SPEED        15.0                           // Maximum speed used to derive deltaZ value
#define HYPERSPACE_LIMIT 3                              // Maximum scale limit for Hyperspace effect

@interface Star : NSObject {

    GLfloat     x, y, z;

    GLfloat     warpSpeed;
    GLfloat     maxDisks;
    GLfloat     maxEdges;
    GLfloat     disk;
    GLfloat     edge;
    GLfloat     deltaZ;
    GLfloat     phi;
    GLfloat     theta;
    GLfloat     rollAngle;
    GLboolean   hyperspace;
    GLint       hyperspaceIndex;
    
    glm::vec4   color;                                  // Color, attenuated by position on ellipsoid
    glm::mat4   transform;                              // Star's position & rotation on 3D ellipsoid
}

-(id)           init			:(int)in_WarpSpeed  :(Vector3f*)in_AxisLimits;
-(void)			reset			:(bool)in_InitAll  :(bool)in_HyperspaceEffect;

-(GLfloat)      getX;
-(GLfloat)      getY;
-(GLfloat)      getZ;
-(GLfloat)      getDisk;
-(GLfloat)      getEdge;
-(GLfloat)      getDeltaZ;
-(GLfloat)      getPhi;
-(GLfloat)      getTheta;
-(GLfloat)      getRollAngle;
-(GLfloat)      getMaxDisks;
-(GLfloat)      getMaxEdges;
-(GLboolean)    getHyperspace;
-(glm::mat4)    getTransform;
-(glm::vec4)    getColor;

-(void)         setX                :(GLfloat)newx;
-(void)         setY                :(GLfloat)newy;
-(void)         setZ                :(GLfloat)newz;
-(void)         setPhi              :(GLfloat)newphi;
-(void)         setTheta            :(GLfloat)newtheta;
-(void)         setDisk			    :(GLfloat)newdisk;
-(void)			setEdge			    :(GLfloat)newedge;
-(void)         setColor            :(glm::vec4)newColor;
-(void)         setTransform        :(glm::mat4)newTransform;
-(void)         setWarpSpeed        :(GLfloat)newWarpSpeed;
-(void)         setHyperspace       :(GLboolean)newHyperspace;

-(GLint)        incHyperspaceIndex;

@end
