# DisplayP2

DisplayP2 for macOS displays and animates P-Squared (.p2) mesh files.

P2 meshes are derived from, and an incompatible superset of, Autodesk's ancient .3ds format.

P2 meshes begin life as Blender models exported to .p2 format using my Blender add-on: [io\_scene\_p2](https://p-squared.com/sources/io_scene_p2.zip).

### Current limitations:

* P2 meshes exported by io\_scene\_p2 are limited to 65,535 vertices and 512 skeletal bones,
* P2 meshes imported by the MeshP2.framework are further limited to about 66 skeletal bones.

### Prerequisites for Users:

* To import P2 meshes into DisplayP2 requires my [MeshP2 Framework](https://p-squared.com/sources/MeshP2-mac.tar.gz) be installed in ***/Library/Frameworks***.

### *Additional* Prerequisites for Developers:

* To compile and build DisplayP2 requires two math library headers be installed prior to building: [GLM](https://github.com/g-truc/glm) and [Eigen](https://eigen.tuxfamily.org/index.php?title=Main_Page). 

* DisplayP2 uses the Superpowered SDK to 'listen' to any ambient audio, analyze its tempo and synchronize animations to the beat:

    [Superpowered SDK - Getting Started](https://docs.superpowered.com/getting-started/how-to-integrate/mac-os?lang=cpp)

DisplayP2 source code is [Free Software](https://www.gnu.org/philosophy/free-sw.en.html) and further placed in the public domain under a [CC0 no copyright declaration](https://creativecommons.org/publicdomain/zero/1.0/).

If you found DisplayP2 instructive, please consider making a charitable donation to the [Free Software Foundation](https://fsf.org).

* * *

Once invoked, roll the mouse-wheel to zoom in and out, or click+hold the mouse-wheel to pan around.

Here's a list of DisplayP2 keystroke commands (all keys are case insensitive):

| Keystroke | Function |
| --- | --- |
| Key_board_ numbers 0 to 9 | Displays some built-in meshes:<br><br>* 1, 2 and 3 Keyboard keys select some cubes with 1, 2 or 3 bones iniitially used for testing. Press the A-key to toggle Armature animation.<br>* 4 selects the original Cowboy mesh from [Karl Wimble's Animation Tutorial.](https://youtu.be/f3Cr8Yx3GGA?si=X-WcvrjITNeVhbHu) Again, press the A-key to toggle Armature animation.<br>* 5 selects the Enterprise mesh. This isn't animated, but looks groovy with the Key_pad_ 1 through 9 Warp-drive effects (below).<br>* 6, 7, 8 and 9 Keyboard keys select the default Bear mesh. Press ⌘-B keys to auto-load its (very cool) .BVH based animation.<br>* 0 selects the 'Loading' mesh, which uses a Blender Action, not an Armature or .BVH animation. Use ⌘-A to toggle Actions. |
| Key_pad_ numbers 1 to 9 | Enables an immersive 3D Warp-drive effect (Keypad 0 to turn off). |
| Keyboard keys |     |
| A Key | Animate mesh's 'Armature' animations, if any. Mutually exclusive with 'Action' animations. |
| ⌘-A Keys | Animate mesh's 'Action' animations, if any. Mutually exclusive with 'Armature' animations. |
| B Key | Animate mesh, after prompting the user for the location of a (compatible) .BVH file. |
| ⌘-B Keys | Animate mesh using .BVH file from same directory as mesh, if available. |
| C Key | 'Correct' mesh with 90° X-axis rotation (simulates Blender rendering). |
| D Key | 'Double' mesh size by moving the camera closer to the mesh. |
| F Key | 'Fly-by' to (slowly) view the mesh from several angles. |
| H Key | 'Halve' mesh size by moving the camera further away. |
| L Key | 'Listen' to any ambient music, synchronize animation to tempo. |
| M Key | Select default Sky 'Map' (i.e. background). |
| ⌘-M Keys | Select Deep Field Sky Map. |
| O Key | Toggles 'Ocean' wave effect. _Warning:_ GPU intensive! |
| R Key | Resets camera, lighting, projections and viewing frustum. |
| S Key | Toggles the currently selected Sky Map on/off. |
| W Key | Toggles wireframe mode on/off for all polygons. |
| ⌘-W Keys | Toggles Animated Warp Field effect on/off. _Warning:_ GPU intensive! |
| ESC, Q or X Keys | Exits the program at top of event loop. |

* * *
