#define GL_DO_NOT_WARN_IF_MULTI_GL_VERSION_HEADERS_INCLUDED

#import <Cocoa/Cocoa.h>
#import <OpenGL/OpenGL.h>
#import <MeshP2/MeshP2.h>
#import <glm/gtx/euler_angles.hpp>

#import "Okudagram.h"
#import "BVH_Parser/bvh.h"

#define FOV_DEGREES         45.0                // (Degrees) Field of View
#define COSINE_THRESHOLD	0.0	                // (Cosine) Threshold angle between surfaces sharing a vertex; decides when mesh vertex is split
#define FORCE_2_SIDED       false               // (BOOL) flag passed through to MeshP2 when mesh is loaded, either initially or on file open
#define SS_THRESHOLD        1000000             // (Bytes) Mesh file size threshold beyond which we display a "Loading..." splash screen
#define SS_DELAY            3.25                // (Seconds) Duration to display animated splash screen before loading requested mesh

@interface View : NSOpenGLView {
    
    MeshP2     *mesh;
    Okudagram  *okudagram;
    glm::vec3   cameraPosition;
    
    GLuint      flyBy_Index;
    bool        flyBy_Enabled;
    bool        P2_Logging_Enabled;

    NSRect		sceneBounds;
    GLboolean	arcballEnabled;
    CGFloat		lastMouseX, lastMouseY, currentMouseX, currentMouseY;
    
    Matrix4f   *modelViewMatrix;
    Matrix4f   *projectionMatrix;
    
    Vector4f   *lightPosition;
    Vector4f   *lightAmbient;
    Vector4f   *lightDiffuse;
    Vector4f   *lightSpecular;
    
    GLfloat     lightConstantAttenuation;
    GLfloat     lightLinearAttenuation;
    GLfloat     lightQuadraticAttenuation;
}

-(id)		  initWithFrame	        :(NSRect)frame  :(bool)in_Logging;
-(id)		  loadP2MeshFromBundle  :(unichar)in_Index;
-(void)       loadP2MeshByFileName  :(NSString*)in_FileName;
-(Animation*) loadBVHanimation      :(NSString*)in_FileName  :(float)in_CurrentBPM;
-(void)       initCameraLightingAndMatrices  :(bool)in_InitEverything;

-(void)	      reshape;
-(void)	      drawRect		:(NSRect)rect;

-(Matrix4f*)  buildModelViewMatrix;
-(Matrix4f*)  buildProjectionMatrix;

-(MeshP2*)    getMesh;
-(Matrix4f*)  getModelViewMatrix;
-(Okudagram*) getOkudagram;
-(glm::vec3)  getCameraPosition;
-(BOOL)       getFlyBy_Enabled;
-(GLuint)     getFlyBy_Index;

-(void)       setMesh               :(MeshP2*)in_NewMesh;
-(void)       setModelViewMatrix    :(Matrix4f*)in_newModelViewMatrix;
-(void)       setOkudagram          :(Okudagram*)in_newOkudagram;
-(void)       setCameraPosition     :(glm::vec3)in_newCameraPosition;
-(void)       setFlyBy_Enabled      :(BOOL)in_newFlyByEnabled;
-(void)       setFlyBy_Index        :(GLuint)in_newFlyByIndex;

int           signum (int n);

@end
