#import "View.h"

@implementation View

-(id) initWithFrame  :(NSRect)frame  :(bool)in_Logging {
    
    mesh                = nil;
    okudagram           = nil;
    P2_Logging_Enabled  = in_Logging;
    arcballEnabled      = false;
    flyBy_Enabled       = false;
    flyBy_Index         = 0;
    cameraPosition      = glm::vec3 (0.0);
    lastMouseX          = lastMouseY    =   currentMouseX   = currentMouseY     = 0.0;

    NSOpenGLPixelFormatAttribute pixelAttribs[] = { NSOpenGLPFAOpenGLProfile,
                                                    NSOpenGLProfileVersion3_2Core,
                                                    NSOpenGLPFADoubleBuffer,
                                                    NSOpenGLPFAAccelerated,
                                                    NSOpenGLPFAColorSize, 16,
                                                    NSOpenGLPFADepthSize, 24,
                                                    NSOpenGLPFAMultisample,            // The remaining attributes are for texture multisampling (full screen anti-aliasing)
                                                    NSOpenGLPFASampleBuffers, 1,
                                                    NSOpenGLPFASamples, 4,
                                                    NSOpenGLPFANoRecovery,
                                                    0 };
    
    NSOpenGLPixelFormat    *pixelFormat = [[NSOpenGLPixelFormat alloc] initWithAttributes  :pixelAttribs];
    
    if  (pixelFormat != nil) {
        
        self = [super initWithFrame  :frame pixelFormat:pixelFormat];

        if  (self) {
            
            [[self openGLContext] makeCurrentContext];
        }
    }
    
    else self = nil;
    
    return self;
}

-(id) loadP2MeshFromBundle  :(unichar)in_Index {
    
    @try {
        
        NSString *meshFileName = nil;
        NSBundle *bundle       = [NSBundle bundleForClass  :[self class]];

        switch (in_Index) {
                
            case '1':
                
                meshFileName = [bundle pathForResource  :@"1boneCube" ofType:@"p2"];
                break;
                
            case '2':
                
                meshFileName = [bundle pathForResource  :@"2boneCube" ofType:@"p2"];
                break;
                
            case '3':
                
                meshFileName = [bundle pathForResource  :@"3boneCube" ofType:@"p2"];
                break;
                
            case '4':
                
                meshFileName = [bundle pathForResource  :@"Cowboy" ofType:@"p2"];
                break;
                
            case '5':
                
                meshFileName = [bundle pathForResource  :@"NCC-1701-D" ofType:@"p2"];
                break;
                
            case '6':
            case '7':
            case '8':
            case '9':
                
                meshFileName = [bundle pathForResource  :@"Bear" ofType:@"p2"];
                break;

            case '0':
                
                meshFileName = [bundle pathForResource  :@"Loading" ofType:@"p2"];
                break;
                
            default:
                
                meshFileName = [bundle pathForResource  :@"Bear" ofType:@"p2"];
        }

        [self loadP2MeshByFileName  :meshFileName];
    }
    
    @catch    (NSException *ex) {

        NSAlert *alert = [NSAlert new];

        [alert setAlertStyle        :NSAlertStyleWarning];
        [alert setMessageText       :@"DisplayP2 failed to loadMeshFromBundle!"];
        [alert setInformativeText   :@"Error parsing the built-in (i.e. bundled) .p2 mesh resource.\n\nTry opening another .p2 file."];
        [alert addButtonWithTitle   :@"OK"];

        [alert runModal];
        
        NSArray *backtrace = [ex callStackSymbols];
        
        [MeshP2 P2Log  :P2_Logging_Enabled  :@"Exception raised in loadMeshFromBundle e = %@.\n\n Backtrace:\n%@", ex, backtrace];
        
        return nil;
    }

    return    mesh;
}

-(void) loadP2MeshByFileName  :(NSString*)in_FileName {
    
    __block Vector3f *limits;

    NSBundle *bundle = [NSBundle bundleForClass  :[self class]];

    NSString *splashScreenMesh = [bundle pathForResource  :@"Loading"  ofType:@"p2"];
    
    mesh = [[MeshP2 alloc] init  :COSINE_THRESHOLD  :FORCE_2_SIDED  :P2_Logging_Enabled];

    int fileSize = (int)[[[NSFileManager defaultManager] attributesOfItemAtPath  :in_FileName error:nil] fileSize];

    if (fileSize > SS_THRESHOLD) {                                  // If the requested mesh is considered 'large'...

        [mesh ParseP2  :splashScreenMesh];                          // ... show a splash screen before large mesh loads
        
        limits = [[Vector3f new] init  :[mesh getMax_X]  :[mesh getMax_Y]  :[mesh getMax_Z]];

        [self setOkudagram  :[[Okudagram alloc] init  :DEF_SPEED  :DEF_STARS  :limits  :P2_Logging_Enabled]];

        [self initCameraLightingAndMatrices  :true];                // Init the View's camera, lighting, MVPs & view frustum
        
        NSEvent *keyDownEvent = [NSEvent keyEventWithType:NSEventTypeKeyDown
                                                 location:NSZeroPoint
                                            modifierFlags:NSCommandKeyMask  // Send ⌘-a key to play splash screen Action animation
                                                timestamp:NSProcessInfo.processInfo.systemUptime
                                             windowNumber:0
                                                  context:nil
                                               characters:@"a"
                              charactersIgnoringModifiers:@"a"
                                                isARepeat:NO
                                                  keyCode:0];
        
        [NSApp postEvent  :keyDownEvent  atStart:YES];              // Post the keyDown event to animate the splash screen
        
        // Load the large mesh on a background thread (after a short delay to give splash screen animation time to settle)

        dispatch_after (dispatch_time (DISPATCH_TIME_NOW, SS_DELAY * NSEC_PER_SEC), dispatch_get_main_queue(), ^{
            
            [NSApp postEvent  :keyDownEvent  atStart:NO];           // After short delay, turn off Splash screen
            
            [self->mesh ParseP2  :in_FileName];                     // Parse and display the requested mesh
            
            limits = [[Vector3f new] init  :4.0  :4.0  :4.0];       // Experiment with a nominal starfield size

            [self setOkudagram  :[[Okudagram alloc] init  :DEF_SPEED  :DEF_STARS  :limits  :self->P2_Logging_Enabled]];
            
            [self initCameraLightingAndMatrices  :true];            // Init the View and recompute the viewing frustum
        });

    } else {                                                        // Otherwise, parse the requested mesh file immediately
        
        [mesh ParseP2  :in_FileName];
        
        limits = [[Vector3f new] init  :[mesh getMax_X]  :[mesh getMax_Y]  :[mesh getMax_Z]];

        [self setOkudagram  :[[Okudagram alloc] init  :DEF_SPEED  :DEF_STARS  :limits  :self->P2_Logging_Enabled]];

        [self initCameraLightingAndMatrices  :true];                // Init the Views camera, lighting, MVPs, view frustum
    }
}

-(Animation*) loadBVHanimation  :(NSString*)in_FileName  :(float)in_CurrentBPM {
/*
    Unlike Armatures and Actions, which MeshP2 loads from Blender-exported .P2 mesh files,
    .BVH files are processed here and its joint data uploaded to MeshP2 as a new BVHList.
    Lastly, return a new Animation object containing armature-style animation KeyFrames,
    which MeshP2 uses to calculate inter-frame progressions & animation interpolations.
*/
    BVH            *bvh = new BVH();                                    // BVH Parser from https://github.com/sgynn/bvh-browser
    
    uint            jointCount, frameCount;
    bool            parseResult;
    float           frameTime;
    
    NSError        *fileError;
    NSString       *fileContents = [NSString stringWithContentsOfFile  :in_FileName encoding:NSUTF8StringEncoding error:&fileError];

    NSMutableArray *newBVHList            = [NSMutableArray new];       // Array of BVH joints for MeshP2 to animate
    NSMutableArray *newAnimationKeyframes = [NSMutableArray new];       // Array of joint's keyframes to interpolate
    NSMutableArray *newAnimationDurations = [NSMutableArray new];       //   "   "    "     durations "      "

    if (!fileError) parseResult  = bvh->load (fileContents.UTF8String); // Parse BVH file into component parts (joints)
                            
    if (fileError || (!parseResult)) {                                  // Bail if either file load or BVH parsing error
        
        NSString *errorMessage = [NSString stringWithFormat  :@"File Error Description: %@", [fileError localizedDescription]];
        
        NSException* myException = [NSException exceptionWithName  :@"BVHFileException" reason:errorMessage userInfo:nil];

        [MeshP2 P2Log  :P2_Logging_Enabled  :@"DisplayP2->parseBVHByFileName: %@, BVH Parser: %@", errorMessage, (parseResult) ? @"PASS" : @"FAIL"];

        if (bvh) delete bvh;

        [myException raise];
    }

    jointCount  = bvh->getPartCount();                                                      // Number of joints in this BVH animation
    frameCount  = bvh->getFrameCount();                                                     // Number of frames in this BVH animation

    if (in_CurrentBPM) {                                                                    // If the current tempo is detected/valid
        
        frameTime   = ((in_CurrentBPM / 60.0) * (1.0 / 24.0) / 2.0);                        // Match frame time to tempo (half beat)

    } else {
        
        frameTime   = bvh->getFrameTime();                                                  // BVH's time per animation frame (in ms)
    }

    if (jointCount > MAX_BONES) {                                                           // Does the joint count exceed the maximum?

        NSString *errorMessage = [NSString stringWithFormat  :@"BVH file has %d defined bones, exceeding MAX_BONES limit of %d", jointCount, MAX_BONES];
                
        NSException* myException = [NSException exceptionWithName  :@"BVHFileException" reason:errorMessage userInfo:nil];
        
        [MeshP2 P2Log  :P2_Logging_Enabled  :@"%@", errorMessage];

        if (bvh) delete bvh;

        [myException raise];
    }

    for (int jointIdx = 0; jointIdx < jointCount; jointIdx++) {                             // For each part (i.e. joint) in the BVH data:
        
        glm::vec3   vec3Pose;                                                               // Holds joint's current frame position vector
        float       mat4Pose[16];                                                           // Holds joint's current frame rotation matrix
        float       lastFrameTime = 0.0;                                                    // Each joint uses a separate animation timing

        NSMutableArray *jointKeyFrames = [NSMutableArray arrayWithCapacity  :frameCount];   // Holds joint's keyframes (for interpolation)

        const BVH::Part *part = bvh->getPart (jointIdx);                                    // Per BVH_Parser terminology, get part/joint
        
        NSString *jointName = [NSString stringWithUTF8String  :part->name];                 // Get the joint name from BVH_Parser object

        glm::vec3 jointOffset = glm::vec3 (part->offset.x, part->offset.y, part->offset.z); // Get Local-space offset from parent joint

        glm::mat4 lsBindTransform = glm::translate (glm::mat4 (1.0), jointOffset);          // Joint's (Local-space) bind as transform

        for (int frameIdx = 0; frameIdx < frameCount; frameIdx++) {                         // For each of the joint's frames/poses:
            
            lastFrameTime = frameIdx * frameTime;                                           // Timestamp is function of frame time

            BVH_Math::Transform pose = part->motion[frameIdx];                              // Get this frame's pose transform

            vec3Pose = glm::vec3 (pose.offset.x, pose.offset.y, pose.offset.z);             // Get pose transform's position

            pose.toMatrix (mat4Pose);                                                       // Get pose transform's rotation as matrix

            glm::mat4  glmMat4Pose = glm::make_mat4 (mat4Pose);                             // Convert pose matrix to glm::mat4 format
            
            DualQuaternion *dqPose = [[DualQuaternion alloc] initFromGLM  :glmMat4Pose];    // Convert pose to DualQuaternion format

            BoneKeyFrame *keyframe = [[BoneKeyFrame alloc] init  :lastFrameTime  :dqPose];  // Create a timestamped BoneKeyFrame
            
            [jointKeyFrames  addObject  :keyframe];                                         // Add to array of joint's keyframes

            if (P2_Logging_Enabled && (jointIdx == 0) && (frameIdx < MAX_DEBUG_FRAMES)) {   // Read out a few debugging frames

                glm::quat glmQuat     = glm::quat_cast (glmMat4Pose);                       // Get Local-space rotation in glm::quat format
                
                NSString *strMat4Pose = [NSString stringWithUTF8String  :glm::to_string (glmMat4Pose).c_str()];  // Pose transform as matrix string
                NSString *strQuatPose = [NSString stringWithUTF8String  :glm::to_string (glmQuat).c_str()];      // Pose rotation quaternion string

                NSString *strFormat   = [NSString stringWithUTF8String  :"DisplayP2->parseBVHByFileName:\t(\"%@\")\tJoint:%d (\"%@\")\tFrame:%d\n"
                                        "\tColumnMajorMat4:\t\t\t%@\n\tRotationQuaternion:\t\t\t%@\n\tPosition (Local-space):\t\t(%f, %f, %f)\n"];

                [MeshP2 P2Log  :P2_Logging_Enabled  :strFormat, [in_FileName lastPathComponent], jointIdx, jointName, frameIdx,
                                                    strMat4Pose, strQuatPose, vec3Pose.x, vec3Pose.y, vec3Pose.z];
            }
        }
            
        assert (jointIdx == part->index);                                                   // Extra continuity check to avoid traumas
        
        BONE   *bvhJoint = [[BONE alloc] init  :jointIdx  :jointName  :part->parent];       // Instantiate a joint for the new BVHList
        
        NSMutableArray *childJoints = [NSMutableArray new];                                 // Array of Child BVH joints MeshP2 will loop through

        std::vector<int> childIndicesVector = part->childIndices;                           // Get this joint's vector array of child joint indices

        std::for_each (childIndicesVector.begin(), childIndicesVector.end(), ^(int childIndex) {
            
            [childJoints addObject  :[[NSNumber alloc] initWithInt  :childIndex]];          // Convert child joint indices to NSMutableArray
        });
        
        [bvhJoint setLsBindTransform  :[[DualQuaternion alloc] initFromGLM  :lsBindTransform]]; // Joint's (Local-space) Bind Transform

        [bvhJoint    setChildIndices  :childJoints];                                        // Copy Child Indices to target joint for newBVHList
        
        newBVHList[jointIdx] = bvhJoint;                                                    // Insert in new BVHList at the appropriate bone index
        
        [newAnimationKeyframes addObject  :jointKeyFrames];                                 // Add keyframes to array of joint's keyframes; see (2)
        
        [newAnimationDurations addObject  :[NSNumber numberWithFloat  :lastFrameTime]];     // Add duration to array of joint's animation durations
    }

    if (bvh) delete bvh;                                                                    // All done with BVH Parser object; prevent memory leak
    
    if ([newBVHList count]) [[self getMesh] UploadBVHList  :newBVHList];                    // Send the new BVHList, if any, to our MeshP2 instance

    [MeshP2 P2Log  :P2_Logging_Enabled  :@"DisplayP2->parseBVHByFileName:\t(\"%@\"), jointCount:%d, frameCount:%d, frameTime:%f",
                                        [in_FileName lastPathComponent], jointCount, frameCount, frameTime];
    
    return [[Animation alloc] init  :newAnimationKeyframes  :newAnimationDurations];        // Return a new (and possibly empty) 'Animation' object
}

-(void) initCameraLightingAndMatrices  :(bool)in_InitEverything {
    
    if (in_InitEverything) {                                                                // Initialize everything, or just the matrices?
        
        [MeshP2 P2Log  :P2_Logging_Enabled  :@"DisplayP2->initCameraLightingAndMatrices: Mesh Axis Maximums: (%.03f, %.03f, %.03f)",
                                            [mesh getMax_X], [mesh getMax_Y], [mesh getMax_Z]];

        float limit = std::max ({ [mesh getMax_X], [mesh getMax_Y], [mesh getMax_Z] });     // Find highest value among mesh's X, Y and Z axes to position camera

        cameraPosition  = glm::vec3 (0.0, 0.0, (limit * 3.0));                              // Reset View's default camera position relative to incoming mesh size
    
        lightPosition   = [[Vector4f alloc] init  :0.0f :0.0f :cameraPosition.z W:0.0f];    // W component of 0.0 denotes a Directional Light, and (x, y, z) coordinates...
                                                                                            // ... is the direction vector along which light rays are cast, towards origin
        lightAmbient    = [[Vector4f alloc] init  :0.3f :0.3f :0.3f W:1.0f];
        lightDiffuse    = [[Vector4f alloc] init  :1.0f :1.0f :1.0f W:1.0f];
        lightSpecular   = [[Vector4f alloc] init  :1.0f :1.0f :1.0f W:1.0f];
        
        lightConstantAttenuation  = 1.0f;
        lightLinearAttenuation    = 0.0f;
        lightQuadraticAttenuation = 0.0f;
        
        [mesh SetLightProperties  :lightPosition                                            // Set mesh's global lights from view's
                                  :lightAmbient  :lightDiffuse  :lightSpecular
                                  :lightConstantAttenuation  :lightLinearAttenuation  :lightQuadraticAttenuation];
    }

    modelViewMatrix = [self buildModelViewMatrix];                                          // Build the default ModelView matrix
    projectionMatrix= [self buildProjectionMatrix];                                         // Build the default Projection matrix

    [mesh      SetModelViewMatrix  :modelViewMatrix];                                       // Set the mesh's MVP matrices
    [mesh      SetProjectionMatrix :projectionMatrix];

    [okudagram SetModelViewMatrix  :modelViewMatrix];                                       // Set Okudagram's MVP matrices...
    [okudagram SetProjectionMatrix :projectionMatrix];                                      // ... and when MVP matrices are set...
    [okudagram ExtractFrustum];                                                             // ... reset view frustum plane equations
}

-(Matrix4f*) buildModelViewMatrix {

    glm::mat4 lookAtMatrix = glm::lookAt (glm::vec3 (cameraPosition.x, cameraPosition.y, cameraPosition.z), // Camera (i.e. Eye) position
                                          glm::vec3 (0, 0, 0),                                              // Point being looked at
                                          glm::vec3 (0, 1, 0));                                             // Camera's UP vector
        
    return [[Matrix4f new] cmMatrixFromBytes  :glm::value_ptr (lookAtMatrix)];              // Return new Column-Major ModelViewMatrix
}

-(Matrix4f*) buildProjectionMatrix {

    sceneBounds = [self bounds];

    glViewport (0, 0, sceneBounds.size.width, sceneBounds.size.height);

    glm::mat4 perspectiveMatrix = glm::perspective (glm::radians (FOV_DEGREES),                             // Perspective Field-of-View
                                                    sceneBounds.size.width / sceneBounds.size.height,       // Viewing area's aspect ratio
                                                    NEAR_CLIPPING_PLANE,                                    // Near clipping plane distance
                                                    FAR_CLIPPING_PLANE);                                    // Far  clipping plane distance

    return [[Matrix4f new] cmMatrixFromBytes  :glm::value_ptr (perspectiveMatrix)];         // Return new Column-Major ProjectionMatrix
}

-(MeshP2*) getMesh { return mesh; }

-(Matrix4f*) getModelViewMatrix { return modelViewMatrix; }

-(Okudagram*) getOkudagram { return okudagram; }

-(glm::vec3) getCameraPosition { return cameraPosition; }

-(BOOL) getFlyBy_Enabled { return flyBy_Enabled; }

-(GLuint) getFlyBy_Index { return flyBy_Index; }

-(void) setMesh  :(MeshP2*)in_NewMesh { mesh = in_NewMesh; }

-(void) setModelViewMatrix  :(Matrix4f*)in_newModelViewMatrix { modelViewMatrix = in_newModelViewMatrix; }

-(void) setOkudagram  :(Okudagram*)in_newOkudagram { okudagram = in_newOkudagram; }

-(void) setCameraPosition  :(glm::vec3)in_newCameraPosition { cameraPosition = in_newCameraPosition; }

-(void) setFlyBy_Enabled  :(BOOL)in_newFlyByEnabled { flyBy_Enabled = in_newFlyByEnabled; }

-(void) setFlyBy_Index  :(GLuint)in_newFlyByIndex { flyBy_Index = in_newFlyByIndex; }

-(void) reshape {
    
    [[self openGLContext] update];
    
    [super reshape];
}

-(void) drawRect  :(NSRect)rect {

    glClear  (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);                               // Clear color and depth values from last frame
    
    glEnable (GL_DEPTH_TEST);                                                           // Discard fragments if their Z values fail depth test
   
    if (mesh && flyBy_Enabled) {
        
        if (flyBy_Index == FAR_CLIPPING_PLANE) flyBy_Index = 0;
        
        else {
            
            GLfloat phi     = (GLfloat) (flyBy_Index * M_PI   / FAR_CLIPPING_PLANE);    // Calculate Z-axis position  (pole-to-pole)
            GLfloat theta   = (GLfloat) (flyBy_Index * M_PI*2 / FAR_CLIPPING_PLANE);    // Calculate X and Y position (side-to-side)

            glm::mat4 flybyMatrix = glm::lookAt (glm::vec3 (glm::min ((float)FAR_CLIPPING_PLANE, cameraPosition.x * sin (phi) * cos (theta)),
                                                            glm::min ((float)FAR_CLIPPING_PLANE, cameraPosition.y * sin (phi) * sin (theta)),
                                                            glm::min ((float)FAR_CLIPPING_PLANE, cameraPosition.z * cos (phi))), // Camera's new position
                                                 glm::vec3 (0, 0, 0),                                                            // Point being looked at
                                                 glm::vec3 (0, 1, 0));                                                           // Camera's UP direction

            modelViewMatrix = [[Matrix4f new] cmMatrixFromBytes  :glm::value_ptr (flybyMatrix)];    // Update glView's modelView matrix
            
            [mesh      SetModelViewMatrix  :modelViewMatrix];                                       // Set updated modelView matrix on mesh
            [okudagram SetModelViewMatrix  :modelViewMatrix];                                       // Set updated modelView matrix on okudagram
            [okudagram ExtractFrustum];                                                             // Update Okudagram's view frustum plane equations
        }
        
        flyBy_Index++;
    }

    if (mesh) [mesh DisplayMesh];
    
    if (okudagram) [okudagram Display];
    
    [[self openGLContext] flushBuffer];
}

-(void) otherMouseDown  :(NSEvent*)pEvent {                                 // Use the 3rd mouse button for arcball movement, à la Blender
    
    NSPoint pointInWindow       = [pEvent locationInWindow];
    NSPoint pointInView         = [self convertPoint:pointInWindow fromView:nil];
    
    lastMouseX = currentMouseX  = pointInView.x;
    lastMouseY = currentMouseY  = pointInView.y;
    
    arcballEnabled = true;
    
    [self setNeedsDisplay  :YES];
}

-(void) otherMouseDragged  :(NSEvent*)pEvent {
    
    if (mesh && arcballEnabled) {
        
        NSPoint pointInWindow   = [pEvent locationInWindow];
        NSPoint pointInView     = [self convertPoint:pointInWindow fromView:nil];

        currentMouseX           = pointInView.x;
        currentMouseY           = pointInView.y;
        
        glm::vec2 mouseCoords   = glm::vec2 (pointInView.x/sceneBounds.size.width, pointInView.y/sceneBounds.size.height);
        
        [okudagram SetMouseCoords  :mouseCoords];                                                   // Normalized mouse coordinates
        
        glm::mat4x4 rotationMatrixX (1.0f);
        glm::mat4x4 rotationMatrixY (1.0f);
        
        glm::vec4   position        (cameraPosition.x, cameraPosition.y, cameraPosition.z, 1);      // Homogenous position of camera
        glm::vec4   pivot           (0, 0, 0, 1);                                                   // Homogenous position of pivot (i.e. origin)

        glm::vec3   right_vector    = glm::vec3 (1, 0,  0);
        glm::vec3   upwards_vector  = glm::vec3 (0, 1,  0);
        glm::vec3   view_direction  = glm::vec3 (0, 0, -1);
        
        if  (currentMouseX != lastMouseX || currentMouseY != lastMouseY) {

            // Step 1 : Determine rotation from mouse movement (See example code at https://asliceofrendering.com/camera/2019/11/30/ArcballCamera/)
            
            float deltaAngleX   = (M_PI * 2 / sceneBounds.size.width);                              // Movement from left to right = 2*PI = 360 deg
            float deltaAngleY   = (M_PI     / sceneBounds.size.height);                             // Movement from top to bottom =   PI = 180 deg

            float cosAngle      = dot (view_direction, upwards_vector);                             // Is the view direction aligned straight up, or down?
            
            if ((cosAngle * signum (deltaAngleY)) > 0.99f) deltaAngleY = 0;                         // Handle case where camera direction equals up vector
            
            float xAngle        = (lastMouseX - currentMouseX) * deltaAngleX;
            float yAngle        = (lastMouseY - currentMouseY) * deltaAngleY;
            
            rotationMatrixX     = glm::rotate (rotationMatrixX, xAngle, upwards_vector);            // Step 2: Rotate camera around first axis pivot point
            
            position            = (rotationMatrixX * (position - pivot)) + pivot;

            rotationMatrixY     = glm::rotate (rotationMatrixY, yAngle, right_vector);              // Step 3: Rotate camera around second axis pivot point

            cameraPosition      = glm::vec3 ((rotationMatrixY * (position - pivot)) + pivot);       // Save updated cameraPosition in synthesized ivar
            
            [self initCameraLightingAndMatrices  :false];                                           // Reset lighting & matrices, but leave camera

            lastMouseX = currentMouseX;
            lastMouseY = currentMouseY;
        }

        [self setNeedsDisplay  :YES];
    }
}

-(void) otherMouseUp  :(NSEvent*)pEvent {
    
    arcballEnabled = false;
    
    [self setNeedsDisplay  :YES];
}

-(void) scrollWheel  :(NSEvent*)pEvent {
    
    if    (([pEvent deltaY] != 0.0) && (mesh)) {

        float scaleFactor = ([pEvent deltaY] > 0.0) ? 0.9 : 1.1;            // Scroll wheel up/down equals zoom in/out

        glm::vec3 ncp = cameraPosition * scaleFactor;                       // Calculate potential New Camera Position

        float limit = std::max ({abs(ncp.x), abs(ncp.y), abs(ncp.z)});      // Camera's highest value seen on any axis

        if (limit < FAR_CLIPPING_PLANE) {                                   // Ensure it is within the viewing frustum

            cameraPosition = ncp;                                           // If so, lock it in. Otherwise, ignore it
            
            [self initCameraLightingAndMatrices  :false];                   // Reset lighting/matrices; use new camera
        }
        
        [self setNeedsDisplay  :YES];
    }
}

int signum (int n) { return (n < 0) ? -1 : (n > 0) ? +1 : 0; }

@end

//  (1) Benjamin Kenwright: Dual Quaternions - From Classical Mechanics to Computer Graphics and Beyond
//      (https://xbdev.net/misc_demos/demos/dual_quaternions_beyond/paper.pdf)
//
//  (2) Bone/Joint 'animations' array .vs. 'Animation' object:
//
//      2.1: Bone/Joint 'animations' array: - Only for Blender-imported Armatures and Actions; doesn't apply to BVH files.
//                                          - An array of BONE object's poses in Matrix4f format, each item being 1 frame.
//                                          - Set by MeshP2->Convert_to_OpenGL to populate ArmatureList animation entries.
//                                          - Set by MeshP2->Convert_to_OpenGL to populate ActionList (synthesized) animation.
//                                          - DisplayP2->Controller->loadAnimationByType will use Armature/ActionList to build
//                                            an 'Animation' object, which Controller->keyDown slots in Controller->animations
//                                            and uploads to MeshP2 with a call to the MeshP2->UploadAnimation instance method; see 2.2.
//
//                                          - BVH files, however, aren't Blender-imported, so DisplayP2->parseBVHByFileName instead:
//                                              - Creates a BVHList array like MeshP2->Convert_to_OpenGL did for Armature/ActionList
//                                              - For each joint in BVHList, sets the (Local-space) Bind Transform and Child Indices
//                                                (all other BONE items, like an 'animations' entry, use default values (e.g. 'nil')
//                                              - Uploads the new BVHList by calling the MeshP2->UploadBVHList instance method, then
//                                              - Returns an 'Animation' object using only the keyframes parsed from the BVH file,
//                                                which Controller->keyDown subsequently slots in the Controller->animations array
//                                                and uploads to MeshP2 with a call to the MeshP2->UploadAnimation instance method; see 2.2.
//
//      2.2: 'Animation' object:            - Uploaded to MeshP2 when DisplayP2 calls the MeshP2->UploadAnimation instance method.
//                                          - Then, MeshP2->[Animator calculateCurrentAnimationPoses] uses the 'Animation' object:
//                                              - For each bone in the 'Animation' object:
//                                                  - Find neighbouring keyframes by timestamps,
//                                                  - Compute current interpolation progression,
//                                                  - Compute interpolated transform at progression as DualQuaternion,
//                                                  - Upload interpolated transform to a GPU shader via loadBoneTransformUniforms.
//
//      In essence, the Bone/Joint 'animations' array is an intermediate form of 'Animation' keyframes, which ultimately defines animations.
//
//  (3) BVH files may not have positions for all joints as it's a Blender option to "Only write out translation channels for the root bone".
//
